var jotai = (function (exports, react, scheduler, useContextSelector) {
  'use strict';

  function _extends() {
    _extends = Object.assign || function (target) {
      for (var i = 1; i < arguments.length; i++) {
        var source = arguments[i];

        for (var key in source) {
          if (Object.prototype.hasOwnProperty.call(source, key)) {
            target[key] = source[key];
          }
        }
      }

      return target;
    };

    return _extends.apply(this, arguments);
  }

  function _unsupportedIterableToArray(o, minLen) {
    if (!o) return;
    if (typeof o === "string") return _arrayLikeToArray(o, minLen);
    var n = Object.prototype.toString.call(o).slice(8, -1);
    if (n === "Object" && o.constructor) n = o.constructor.name;
    if (n === "Map" || n === "Set") return Array.from(o);
    if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
  }

  function _arrayLikeToArray(arr, len) {
    if (len == null || len > arr.length) len = arr.length;

    for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

    return arr2;
  }

  function _createForOfIteratorHelperLoose(o, allowArrayLike) {
    var it;

    if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) {
      if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
        if (it) o = it;
        var i = 0;
        return function () {
          if (i >= o.length) return {
            done: true
          };
          return {
            done: false,
            value: o[i++]
          };
        };
      }

      throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
    }

    it = o[Symbol.iterator]();
    return it.next.bind(it);
  }

  var isSSR = typeof window === 'undefined' || /ServerSideRendering/.test(window.navigator && window.navigator.userAgent);
  var useIsoLayoutEffect = isSSR ? function (fn) {
    return fn();
  } : react.useLayoutEffect;

  var MAX_CHAIN_LENGTH = 10;
  var SMALL_MAP_SIZE = 10;
  var mCreate = function mCreate() {
    return [new Map()];
  };
  var mGet = function mGet(m, k) {
    for (var i = 0; i < m.length; ++i) {
      var v = m[i].get(k);

      if (v) {
        return v;
      }
    }
  };

  var squash = function squash(m) {
    var i = m.length - 1;
    var dst = new Map(m[i]);

    while (i) {
      --i;
      m[i].forEach(function (v, k) {
        dst.set(k, v);
      });
    }

    return [dst];
  };

  var mSet = function mSet(m, k, v) {
    if (m[0].size < SMALL_MAP_SIZE) {
      return [new Map(m[0]).set(k, v)].concat(m.slice(1));
    }

    if (m.length >= MAX_CHAIN_LENGTH) {
      m = squash(m);
    }

    return [new Map([[k, v]])].concat(m);
  };
  var mDel = function mDel(m, k) {
    if (m.length > 1) {
      m = squash(m);
    }

    var map = new Map(m[0]);
    map.delete(k);
    return [map];
  };
  var mKeys = function mKeys(m) {
    var keys = new Set();
    m.forEach(function (map) {
      for (var _iterator = _createForOfIteratorHelperLoose(map.keys()), _step; !(_step = _iterator()).done;) {
        var key = _step.value;
        keys.add(key);
      }
    });
    return keys;
  };
  var mMerge = function mMerge(m, newM) {
    m = [].concat(newM, m);

    if (m.length > MAX_CHAIN_LENGTH) {
      m = squash(m);
    }

    return m;
  };
  var mToPrintable = function mToPrintable(m, toPrintableK, toPrintableV) {
    if (m.length > 1) {
      m = squash(m);
    }

    return new Map([].concat(m[0].entries()).map(function (_ref) {
      var k = _ref[0],
          v = _ref[1];
      return [toPrintableK(k), toPrintableV(v)];
    }));
  };

  var warningObject = new Proxy({}, {
    get: function get() {
      throw new Error('Please use <Provider>');
    },
    apply: function apply() {
      throw new Error('Please use <Provider>');
    }
  });

  var useWeakMapRef = function useWeakMapRef() {
    var ref = react.useRef();

    if (!ref.current) {
      ref.current = new WeakMap();
    }

    return ref.current;
  };

  var warnAtomStateNotFound = function warnAtomStateNotFound(info, atom) {
    console.warn('[Bug] Atom state not found. Please file an issue with repro: ' + info, atom);
  };

  var initialState = mCreate();
  var initialUsedState = mCreate();

  var updateAtomState = function updateAtomState(atom, prevState, partial, prevPromise, isNew) {
    var atomState = mGet(prevState, atom);

    if (!atomState) {
      if (!isNew && process.env.NODE_ENV !== 'production') {
        warnAtomStateNotFound('updateAtomState', atom);
      }

      atomState = {
        deps: new Set()
      };
    }

    if (prevPromise && prevPromise !== atomState.readP) {
      return prevState;
    }

    return mSet(prevState, atom, _extends({}, atomState, partial));
  };

  var addDependent = function addDependent(atom, dependent, prevState) {
    var atomState = mGet(prevState, atom);

    if (atomState) {
      if (!atomState.deps.has(dependent)) {
        var newDeps = new Set(atomState.deps).add(dependent);
        return mSet(prevState, atom, _extends({}, atomState, {
          deps: newDeps
        }));
      }
    } else if (process.env.NODE_ENV !== 'production') {
      warnAtomStateNotFound('addDependent', atom);
    }

    return prevState;
  };

  var replaceDependencies = function replaceDependencies(atom, prevState, dependenciesToReplace) {
    var dependencies = new Set(dependenciesToReplace);
    var nextState = prevState;
    mKeys(nextState).forEach(function (a) {
      var aState = mGet(nextState, a);

      if (aState.deps.has(atom)) {
        if (dependencies.has(a)) {
          // not changed
          dependencies.delete(a);
        } else {
          var newDeps = new Set(aState.deps);
          newDeps.delete(atom);
          nextState = mSet(nextState, a, _extends({}, aState, {
            deps: newDeps
          }));
        }
      }
    });
    dependencies.forEach(function (a) {
      var aState = mGet(nextState, a);

      if (aState) {
        var newDeps = new Set(aState.deps).add(atom);
        nextState = mSet(nextState, a, _extends({}, aState, {
          deps: newDeps
        }));
      } else if (process.env.NODE_ENV !== 'production') {
        warnAtomStateNotFound('replaceDependencies', a);
      }
    });
    return nextState;
  };

  var readAtomState = function readAtomState(atom, prevState, setState, atomStateCache // read state then cache
  ) {
    if (atomStateCache) {
      var _atomState = mGet(prevState, atom);

      if (_atomState) {
        return [_atomState, prevState];
      }

      _atomState = atomStateCache.get(atom);

      if (_atomState) {
        return [_atomState, mSet(prevState, atom, _atomState)];
      }
    }

    var isSync = true;
    var nextState = prevState;
    var error = undefined;
    var promise = undefined;
    var value = undefined;
    var dependencies = new Set();
    var flushDependencies = false;

    try {
      var promiseOrValue = atom.read(function (a) {
        if (dependencies) {
          dependencies.add(a);
        } else {
          setState(function (prev) {
            return addDependent(a, atom, prev);
          });
        }

        if (a !== atom) {
          var _readAtomState = readAtomState(a, nextState, setState, atomStateCache),
              _aState = _readAtomState[0],
              nextNextState = _readAtomState[1];

          if (isSync) {
            nextState = nextNextState;
          } else {
            // XXX is this really correct?
            setState(function (prev) {
              return mMerge(nextNextState, prev);
            });
          }

          if (_aState.readE) {
            throw _aState.readE;
          }

          if (_aState.readP) {
            throw _aState.readP;
          }

          return _aState.value;
        } // a === atom


        var aState = mGet(nextState, a);

        if (aState) {
          if (aState.readP) {
            throw aState.readP;
          }

          return aState.value;
        }

        return a.init; // this should not be undefined
      });

      if (promiseOrValue instanceof Promise) {
        promise = promiseOrValue.then(function (value) {
          var dependenciesToReplace = dependencies;
          dependencies = null;
          setState(function (prev) {
            return updateAtomState(atom, replaceDependencies(atom, prev, dependenciesToReplace), {
              readE: undefined,
              readP: undefined,
              value: value
            }, promise);
          });
        }).catch(function (e) {
          var dependenciesToReplace = dependencies;
          dependencies = null;
          setState(function (prev) {
            return updateAtomState(atom, replaceDependencies(atom, prev, dependenciesToReplace), {
              readE: e instanceof Error ? e : new Error(e),
              readP: undefined
            }, promise);
          });
        });
      } else {
        value = promiseOrValue;
        flushDependencies = true;
      }
    } catch (errorOrPromise) {
      if (errorOrPromise instanceof Promise) {
        promise = errorOrPromise;
      } else if (errorOrPromise instanceof Error) {
        error = errorOrPromise;
      } else {
        error = new Error(errorOrPromise);
      }

      flushDependencies = true;
    }

    nextState = updateAtomState(atom, nextState, {
      readE: error,
      readP: promise,
      value: promise ? atom.init : value
    }, undefined, true);

    if (flushDependencies) {
      nextState = replaceDependencies(atom, nextState, dependencies);
      dependencies = null;
    }

    var atomState = mGet(nextState, atom);
    isSync = false;
    return [atomState, nextState];
  };

  var updateDependentsState = function updateDependentsState(atom, prevState, setState) {
    var atomState = mGet(prevState, atom);

    if (!atomState) {
      if (process.env.NODE_ENV !== 'production') {
        warnAtomStateNotFound('updateDependentsState', atom);
      }

      return prevState;
    }

    var nextState = prevState;
    atomState.deps.forEach(function (dependent) {
      if (dependent === atom || typeof dependent === 'symbol' || !mGet(nextState, dependent)) {
        return;
      }

      var _readAtomState2 = readAtomState(dependent, nextState, setState),
          dependentState = _readAtomState2[0],
          nextNextState = _readAtomState2[1];

      var promise = dependentState.readP;

      if (promise) {
        promise.then(function () {
          setState(function (prev) {
            return updateDependentsState(dependent, prev, setState);
          });
        });
        nextState = nextNextState;
      } else {
        nextState = updateDependentsState(dependent, nextNextState, setState);
      }
    });
    return nextState;
  };

  var readAtom = function readAtom(state, readingAtom, setState, readPendingMap, atomStateCache) {
    var prevState = readPendingMap.get(state) || state;

    var _readAtomState3 = readAtomState(readingAtom, prevState, setState, atomStateCache),
        atomState = _readAtomState3[0],
        nextState = _readAtomState3[1];

    if (nextState !== prevState) {
      readPendingMap.set(state, nextState);
    }

    return atomState;
  };

  var writeAtom = function writeAtom(writingAtom, update, setState, addWriteThunk) {
    var pendingPromises = [];

    var writeAtomState = function writeAtomState(prevState, atom, update) {
      var prevAtomState = mGet(prevState, atom);

      if (prevAtomState && prevAtomState.writeP) {
        var promise = prevAtomState.writeP.then(function () {
          addWriteThunk(function (prev) {
            return writeAtomState(prev, atom, update);
          });
        });
        pendingPromises.push(promise);
        return prevState;
      }

      var nextState = prevState;
      var isSync = true;

      try {
        var promiseOrVoid = atom.write(function (a) {
          var aState = mGet(nextState, a);

          if (!aState) {
            if (process.env.NODE_ENV !== 'production') {
              warnAtomStateNotFound('writeAtomState', a);
            }

            return a.init;
          }

          if (aState.readP && process.env.NODE_ENV !== 'production') {
            // TODO will try to detect this
            console.warn('Reading pending atom state in write operation. We need to detect this and fallback. Please file an issue with repro.', a);
          }

          return aState.value;
        }, function (a, v) {
          if (a === atom) {
            var partialAtomState = {
              readE: undefined,
              readP: undefined,
              value: v
            };

            if (isSync) {
              nextState = updateDependentsState(a, updateAtomState(a, nextState, partialAtomState), setState);
            } else {
              setState(function (prev) {
                return updateDependentsState(a, updateAtomState(a, prev, partialAtomState), setState);
              });
            }
          } else {
            if (isSync) {
              nextState = writeAtomState(nextState, a, v);
            } else {
              addWriteThunk(function (prev) {
                return writeAtomState(prev, a, v);
              });
            }
          }
        }, update);

        if (promiseOrVoid instanceof Promise) {
          pendingPromises.push(promiseOrVoid);
          nextState = updateAtomState(atom, nextState, {
            writeP: promiseOrVoid.then(function () {
              addWriteThunk(function (prev) {
                return updateAtomState(atom, prev, {
                  writeP: undefined
                });
              });
            })
          });
        }
      } catch (e) {
        if (pendingPromises.length) {
          pendingPromises.push(new Promise(function (_resolve, reject) {
            reject(e);
          }));
        } else {
          throw e;
        }
      }

      isSync = false;
      return nextState;
    };

    addWriteThunk(function (prevState) {
      return writeAtomState(prevState, writingAtom, update);
    });

    if (pendingPromises.length) {
      return new Promise(function (resolve, reject) {
        var loop = function loop() {
          var len = pendingPromises.length;

          if (len === 0) {
            resolve();
          } else {
            Promise.all(pendingPromises).then(function () {
              pendingPromises.splice(0, len);
              loop();
            }).catch(reject);
          }
        };

        loop();
      });
    }
  };

  var runWriteThunk = function runWriteThunk(lastStateRef, pendingStateRef, setState, contextUpdate, writeThunkQueue) {
    while (true) {
      if (!lastStateRef.current || !writeThunkQueue.length) {
        return;
      }

      var thunk = writeThunkQueue.shift();
      var prevState = pendingStateRef.current || lastStateRef.current;
      var nextState = thunk(prevState);

      if (nextState !== prevState) {
        pendingStateRef.current = nextState;
        Promise.resolve().then(function () {
          var pendingState = pendingStateRef.current;

          if (pendingState) {
            pendingStateRef.current = null;
            contextUpdate(function () {
              scheduler.unstable_runWithPriority(scheduler.unstable_UserBlockingPriority, function () {
                setState(pendingState);
              });
            });
          }
        });
      }
    }
  };

  var ActionsContext = useContextSelector.createContext(warningObject);
  var StateContext = useContextSelector.createContext(warningObject);

  var InnerProvider = function InnerProvider(_ref) {
    var r = _ref.r,
        children = _ref.children;
    var contextUpdate = useContextSelector.useContextUpdate(StateContext);

    if (!r.current) {
      r.current = contextUpdate;
    }

    return children;
  };

  var Provider = function Provider(_ref2) {
    var children = _ref2.children;
    var contextUpdateRef = react.useRef();
    var readPendingMap = useWeakMapRef();
    var atomStateCache = useWeakMapRef();

    var _useState = react.useState(initialState),
        state = _useState[0],
        setStateOrig = _useState[1];

    var lastStateRef = react.useRef(null);
    var pendingStateRef = react.useRef(null);

    var setState = function setState(setStateAction) {
      lastStateRef.current = null;
      var pendingState = pendingStateRef.current;

      if (pendingState) {
        pendingStateRef.current = null;
        setStateOrig(pendingState);
      }

      setStateOrig(setStateAction);
    };

    useIsoLayoutEffect(function () {
      var pendingState = readPendingMap.get(state);

      if (pendingState) {
        setState(pendingState);
        return;
      }

      lastStateRef.current = state;
    });

    var _useState2 = react.useState(initialUsedState),
        used = _useState2[0],
        setUsed = _useState2[1];

    react.useEffect(function () {
      var lastState = lastStateRef.current;
      if (!lastState) return;
      var nextState = lastState;
      var deleted;

      do {
        deleted = false;
        mKeys(nextState).forEach(function (a) {
          var _mGet;

          var aState = mGet(nextState, a); // do not delete while promises are not resolved

          if (aState.writeP || aState.readP) return;
          var depsSize = aState.deps.size;
          var isEmpty = (depsSize === 0 || depsSize === 1 && aState.deps.has(a)) && !((_mGet = mGet(used, a)) == null ? void 0 : _mGet.size);

          if (isEmpty) {
            atomStateCache.set(a, aState);
            nextState = mDel(nextState, a);
            deleted = true;
          }
        });
      } while (deleted);

      if (nextState !== lastState) {
        setState(nextState);
      }
    }, [used, atomStateCache]);
    var writeThunkQueueRef = react.useRef([]);
    react.useEffect(function () {
      runWriteThunk(lastStateRef, pendingStateRef, setState, contextUpdateRef.current, writeThunkQueueRef.current);
    }, [state]);
    var actions = react.useMemo(function () {
      return {
        add: function add(id, atom) {
          setUsed(function (prev) {
            return mSet(prev, atom, new Set(mGet(prev, atom)).add(id));
          });
        },
        del: function del(id, atom) {
          setUsed(function (prev) {
            var oldSet = mGet(prev, atom);
            if (!oldSet) return prev;
            var newSet = new Set(oldSet);
            newSet.delete(id);

            if (newSet.size) {
              return mSet(prev, atom, newSet);
            }

            return mDel(prev, atom);
          });
        },
        read: function read(state, atom) {
          return readAtom(state, atom, setState, readPendingMap, atomStateCache);
        },
        write: function write(atom, update) {
          return writeAtom(atom, update, setState, function (thunk) {
            writeThunkQueueRef.current.push(thunk);

            if (lastStateRef.current) {
              runWriteThunk(lastStateRef, pendingStateRef, setState, contextUpdateRef.current, writeThunkQueueRef.current);
            } else {
              // force update (FIXME this is a workaround for now)
              setState(function (prev) {
                return mMerge(prev, mCreate());
              });
            }
          });
        }
      };
    }, [readPendingMap, atomStateCache]);

    if (process.env.NODE_ENV !== 'production') {
      // eslint-disable-next-line react-hooks/rules-of-hooks
      useDebugState(state);
    }

    return react.createElement(ActionsContext.Provider, {
      value: actions
    }, react.createElement(StateContext.Provider, {
      value: state
    }, react.createElement(InnerProvider, {
      r: contextUpdateRef
    }, children)));
  };

  var atomToPrintable = function atomToPrintable(atom) {
    var _atom$debugLabel;

    return atom.key + ":" + ((_atom$debugLabel = atom.debugLabel) != null ? _atom$debugLabel : '<no debugLabel>');
  };

  var stateToPrintable = function stateToPrintable(state) {
    return mToPrintable(state, atomToPrintable, function (v) {
      return {
        value: v.readE || v.readP || v.writeP || v.value,
        deps: Array.from(v.deps).map(atomToPrintable)
      };
    });
  };

  var useDebugState = function useDebugState(state) {
    react.useDebugValue(state, stateToPrintable);
  };

  var keyCount = 0; // global key count for all atoms

  function atom(read, write) {
    var config = {
      key: ++keyCount
    };

    if (typeof read === 'function') {
      config.read = read;
    } else {
      config.init = read;

      config.read = function (get) {
        return get(config);
      };

      config.write = function (get, set, update) {
        set(config, typeof update === 'function' ? update(get(config)) : update);
      };
    }

    if (write) {
      config.write = write;
    }

    return config;
  }

  var isWritable = function isWritable(atom) {
    return !!atom.write;
  };

  function useAtom(atom) {
    var actions = useContextSelector.useContext(ActionsContext);
    var value = useContextSelector.useContextSelector(StateContext, react.useCallback(function (state) {
      var atomState = actions.read(state, atom);

      if (atomState.readE) {
        throw atomState.readE; // read error
      }

      if (atomState.readP) {
        throw atomState.readP; // read promise
      }

      if (atomState.writeP) {
        throw atomState.writeP; // write promise
      }

      return atomState.value;
    }, [atom, actions]));
    react.useEffect(function () {
      var id = Symbol();
      actions.add(id, atom);
      return function () {
        actions.del(id, atom);
      };
    }, [actions, atom]);
    var setAtom = react.useCallback(function (update) {
      if (isWritable(atom)) {
        return actions.write(atom, update);
      } else {
        throw new Error('not writable atom');
      }
    }, [atom, actions]);
    react.useDebugValue(value);
    return [value, setAtom];
  }

  var useBridge = function useBridge() {
    var actions = useContextSelector.useBridgeValue(ActionsContext);
    var state = useContextSelector.useBridgeValue(StateContext);
    return react.useMemo(function () {
      return [actions, state];
    }, [actions, state]);
  };
  var Bridge = function Bridge(_ref) {
    var value = _ref.value,
        children = _ref.children;
    var actions = value[0],
        state = value[1];
    return react.createElement(useContextSelector.BridgeProvider, {
      context: ActionsContext,
      value: actions
    }, react.createElement(useContextSelector.BridgeProvider, {
      context: StateContext,
      value: state
    }, children));
  };

  exports.Bridge = Bridge;
  exports.Provider = Provider;
  exports.atom = atom;
  exports.useAtom = useAtom;
  exports.useBridge = useBridge;

  Object.defineProperty(exports, '__esModule', { value: true });

  return exports;

}({}, React, scheduler, useContextSelector));
